;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Testing AspectScheme*
;;
;;

#lang scheme
(require "aspectscheme-levels.ss")
(require "test.ss")

(define (level n) (lambda (l) (eq? l n)))

;; write and map are primitives -> no jp
(define (do-write s) (write s))
(define (my-map proc list)
  (if (empty? list)
      list
      (cons (proc (car list)) (my-map proc (cdr list)))))

;;;;;; 
;; a level 0 aspect (writes 0 to stream)

;; a) no use of up/down in (*)

(let ((trace 
       (lambda (proceed) ;; proceed
         (lambda ctx ;; ctx
           (lambda (a) ;; val
             (do-write 'a) ;; (*) ;; writes which advice is executing
             (proceed a))))))
  
  ;; dynamically-scoped level 0
  (test/str
   (fluid-around (&& (call do-write) 
                     (lambda jp (do-write 'pc)'())) ;; (*)
                 trace (level 0)
                 (do-write 'hola))
   "pcahola"))
  ;; pc, adv, base


;;;;;; 
;; a level 1 aspect (writes 1 to stream)
;; a level 0 aspect (writes 0 to stream)

;; a) no use of up/down in (*)

(let ((trace (lambda (mark)
               (lambda (proceed) ;; proceed
                 (lambda ctx ;; ctx
                   (lambda (a) ;; val
                     (do-write mark) ;; (*) ;; writes which advice is executing
                     (proceed a)))))))

;; level 1
  (fluid-around (call do-write) (trace 1) (level 1)
                
    ;; dynamically-scoped level 0
    (test/str
     (fluid-around (&& (call do-write) 
                       (lambda jp (do-write 'pc)'())) ;; (*)
                   (trace 0) (level 0)
                   (do-write 'hola))
     "1pc10hola")))
  ;; adv 1, pc, adv 1, adv 0, base


;; b) move advice/pointcut writes to upper level

(let ((trace (lambda (mark)
               (lambda (proceed) ;; proceed
                 (lambda ctx ;; ctx
                   (lambda (a) ;; val
                     (^up (do-write mark)) ;; (*) ;; writes which advice is executing
                     (proceed a)))))))

  ;; level 1
  (fluid-around (call do-write) (trace 1) (level 1)

  ;; dynamically-scoped level 0
    (test/str
     (fluid-around (&& (call do-write) 
                       (lambda jp (^up (do-write 'pc))'())) ;; (*)
                   (trace 0) (level 0)
                   (do-write 'hola))
     "pc0hola")))
  ;; pc, adv 0, base [adv1 does not apply, since all is up'ed to l2]

;; Testing "invisibility" of aspect activities
(let ((jp-counter (let ((count 0))
                    (lambda ()
                      (lambda (proceed)
                        (if (eq? 'get-count proceed) ;; access the counter
                            count
                            (lambda ctx
                              (lambda args
                                (set! count (+ 1 count))  ;; just count jps
                                (apply proceed args)))))))))
  (let ((counter1 (jp-counter))
        (counter2 (jp-counter)))
    ;;jp counter on level 0
    (fluid-around  (lambda (jp+ jp jp-) '())    ;; true pc
                   counter1
                   (level 0)
      (my-map (lambda (x) (+ x 1)) '(1 2 3 4 5)))
   
    ;; now same thing, with an aspect that stays at its level
    (fluid-around (lambda (jp+ jp jp-) (my-map zero? '(1 2 3)) ;; harmless activity
                    '())
                  (lambda (proceed)
                    (lambda ctx
                      (lambda args
                        (my-map zero? '(1 2 3)) ;; harmless activity
                        (apply proceed args))))
                  (level 0)
      ;;jp counter on level 0
      (fluid-around  (lambda (jp+ jp jp-) '())    ;; true pc
                     counter2
                     (level 0)
        (my-map (lambda (x) (+ x 1)) '(1 2 3 4 5))))
    
    ;; Introducing "pure" aspect should not affect number of join points observable at level 0
    (test (- (counter1 'get-count) (counter2 'get-count)) 0)))


;;;;;;;;;;;; testing level capturing function

;;; Case 1: lambda-s at the same level, so aspect sees it (ie. applies twice)
(let ((trace 
       (lambda (proceed) ;; proceed
         (lambda ctx ;; ctx
           (lambda (a) ;; val
             (do-write '-apply-)
             (proceed a))))))
  
  ;; dynamically-scoped level 0
  (test/str
   (fluid-around (call do-write) 
                 trace (level 0)
                 (do-write 'hola)
                 (let ((do-write-task (lambda-s (x) (do-write x))))
                   (do-write-task 'si)))
   "-apply-hola-apply-si"))

;;; Case 2: lambda-s at upper level, so aspect does not see it (ie. applies once)
(let ((trace 
       (lambda (proceed) ;; proceed
         (lambda ctx ;; ctx
           (lambda (a) ;; val
             (do-write '-apply-)
             (proceed a))))))
  
  ;; dynamically-scoped level 0
  (test/str
   (fluid-around (call do-write) 
                 trace (level 0)
                 (do-write 'hola)
                 (let ((do-write-task (^up (lambda-s (x) (do-write x)))))
                   (do-write-task 'si)))
   "-apply-holasi"))

;;; simple simulation of the schedule-task example from the AOSD submission:
(define task (lambda () "nothing"))
(define (schedule-task t) (set! task t))
(define (run-task) (task))

(define activity 
  (lambda (proceed)
    (lambda ctx
      (lambda args
        (schedule-task (lambda-s () (do-write 'logging))) ;; task captures level 1
        (do-write '-a-)
        (apply proceed args)))))

(test/str
 (fluid-around (call do-write)
               activity (level 0)
               (do-write 'hola)
               (run-task))  ;; aspect does not see the do-write here: matches only once
 "-a-holalogging")