#lang scheme

(provide test
         test/pred
         test/exn
         test/str)

;;; TEST FUNCTIONS
;; test that a value is as expected
(define (test val expected)
  (if (equal? val expected) 
      (printf "OK ~s~n" val)
      (printf "BAD ~s (expected: ~s)~n" val expected)))
      
;; test that a value meets a given predicate
(define (test/pred val pred)
  (if (pred val) 
      (printf "OK ~s~n" val)
      (printf "BAD ~s~n" val)))

;; tests that a given exception occurs (comparison is string=? of the error message)
(define-syntax test/exn
  (syntax-rules ()
    ((_ prog expected)
     (with-handlers ((exn:fail? 
                      (lambda (e)
                        (if (string=? (exn-message e) expected)
                            (printf "OK ~s ~n" expected)
                            (printf "BAD ~s (expected: ~s)~n" 
                                    (exn-message e) expected)))))
       (let ((prog-val prog))
         (printf "BAD ~s (expected: ~s)~n" prog-val expected))))))

;; tests the output of a program on the output-port 
;; (redirects the output in a string output port, and at the end checks against the expected)
;; --tests should use write-string to avoid escaping characters--.
(define-syntax test/str 
  (syntax-rules ()
    ((_ prog expected)
     (let ((result (open-output-string)))
       (parameterize ((current-output-port result))
         prog)
       (if (equal? (get-output-string result) expected)
           (printf "OK ~s~n" expected)
           (printf "BAD ~s (expected: ~s)~n" (get-output-string result)  expected))))))
